@@ -42,9 +42,16 @@
 # include <sys/types.h>
 # include <sys/utime.h>
 #else
+# ifndef _POSIX_SOURCE
+#  define _POSIX_SOURCE
+# endif
 # include <sys/types.h>
 # include <utime.h>
 #endif
+#include <sys/stat.h>
+#include <fcntl.h>
+#include <unistd.h>
+#include <errno.h>
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -56,15 +63,15 @@
 
 /* code adapted from transupp.c begins */
 
-/** Lossless image transformations */
+/** Lossless image transformations (according to the Orientation tag) */
 enum JXFORM_CODE {
-  JXFORM_NONE,		/**< no transformation */
+  JXFORM_NONE = 1,	/**< no transformation */
   JXFORM_FLIP_H,	/**< horizontal flip */
+  JXFORM_ROT_180,	/**< 180-degree rotation */
   JXFORM_FLIP_V,	/**< vertical flip */
   JXFORM_TRANSPOSE,	/**< transpose across UL-to-LR axis */
-  JXFORM_TRANSVERSE,	/**< transpose across UR-to-LL axis */
   JXFORM_ROT_90,	/**< 90-degree clockwise rotation */
-  JXFORM_ROT_180,	/**< 180-degree rotation */
+  JXFORM_TRANSVERSE,	/**< transpose across UR-to-LL axis */
   JXFORM_ROT_270	/**< 270-degree clockwise (or 90 ccw) */
 };
 
@@ -807,6 +814,24 @@ jcopy_markers_execute (struct jpeg_decompress_struct* srcinfo,
 }
 /* code adapted from transupp.c ends */
 
+/**
+ * Modify the file timestamp
+ * @param name		file name
+ * @param datetime	modification and access time
+ */
+static void
+touchfile (const char* name,
+	   time_t datetime)
+{
+  struct utimbuf utim;
+  utim.actime = utim.modtime = datetime;
+  if (utime (name, &utim)) {
+    fputs (name, stderr);
+    perror (": utime");
+    /* non-fatal error, don't abort */
+  }
+}
+
 /** Byte order of the Exif data */
 enum endianness {
   LSB,	/**< most significant byte first (Motorola) */
@@ -924,6 +949,7 @@ static const char* exif_message_table[] = {
   "Exif: JpegIFByteCount out of bounds",
   "Exif: duplicate JpegIFOffset",
   "Exif: JpegIFOffset out of bounds",
+  "Exif: cannot write thumbnail file",
 };
 
 /** error codes from the Exif parser */
@@ -956,23 +982,26 @@ enum exif_message {
   EXIF_JPEGIFB_OOB,		/**< JpegIFByteCount out of bounds */
   EXIF_JPEGIFO_DUP,		/**< duplicate JpegIFOffset */
   EXIF_JPEGIFO_OOB,		/**< JpegIFOffset out of bounds */
+  EXIF_IO_THUMBNAIL,		/**< cannot write thumbnail file */
   EXIF_LASTCODE
 };
 
 /** Parse the Exif information
  * @param cinfo		decompression structure
- * @param xform		automatic (-1) or enum JXFORM_CODE
+ * @param xform		automatic (0) or enum JXFORM_CODE
  * @param timestamp	(input) pointer to timestamp to collect:
  * 0=DateTime, 1=DateTimeOriginal, 2=DateTimeDigitized
  *			(output) the timestamp of the image in the Exif data
  * @param rmifd1	flag: remove IFD1 (containing the thumbnail image)
- * @return		operation to perform on the image, -1 if not Exif
+ * @param thumbname	Exif thumbnail file name (optional)
+ * @return		operation to perform on the image, 0 if not Exif
  */
 static enum JXFORM_CODE
 exif_parse (struct jpeg_decompress_struct* cinfo,
-	    int xform,
+	    enum JXFORM_CODE xform,
 	    time_t* timestamp,
-	    int rmifd1)
+	    int rmifd1,
+	    char* thumbname)
 {
   jpeg_saved_marker_ptr marker;
 
@@ -1251,16 +1280,20 @@ exif_parse (struct jpeg_decompress_struct* cinfo,
 	}
       }
 
-#if 0
-      if (thumb_offset && thumb_size) {
+      if (thumb_offset && thumb_size && thumbname) {
+	FILE* f = fopen (thumbname, "wb");
+	int ok;
+	if (!f)
+	  ERREXIT (cinfo, EXIF_IO_THUMBNAIL);
 	if (thumb_offset + thumb_size > exif_length)
 	  thumb_size = exif_length - thumb_offset;
 	if (!exif_data[thumb_offset + thumb_size - 1])
 	  thumb_size--; /* some cameras add a terminating NUL character */
 	/* output the thumbnail */
-	fwrite (exif_data + thumb_offset, 1, thumb_size, stdout);
+	ok = thumb_size == fwrite (exif_data + thumb_offset, 1, thumb_size, f);
+	if (fclose (f) || !ok)
+	  ERREXIT (cinfo, EXIF_IO_THUMBNAIL);
       }
-#endif
 
       /* remove the IFD1 entry (which contains the thumbnail image) */
       if (rmifd1 && ifd1_offset) {
@@ -1278,20 +1311,14 @@ exif_parse (struct jpeg_decompress_struct* cinfo,
 	case 1: *timestamp = datetimeOriginal; break;
 	case 2: *timestamp = datetimeDigitized; break;
 	}
+
+	if (thumb_offset && thumb_size && thumbname)
+	  touchfile (thumbname, *timestamp);
       }
 
-      if (xform < 0 || xform > 7) {
-	switch (orientation) {
-	default: xform = JXFORM_NONE; break;
-	case 2: xform = JXFORM_FLIP_H; break;
-	case 3: xform = JXFORM_ROT_180; break;
-	case 4: xform = JXFORM_FLIP_V; break;
-	case 5: xform = JXFORM_TRANSPOSE; break;
-	case 6: xform = JXFORM_ROT_90; break;
-	case 7: xform = JXFORM_TRANSVERSE; break;
-	case 8: xform = JXFORM_ROT_270; break;
-	}
-      }
+      if (xform < 1 || xform > 8)
+	if (!(xform = orientation))
+	  xform = JXFORM_NONE;
 
       /* adjust ExifImageWidth and ExifImageHeight */
       if (width_ptr && height_ptr) {
@@ -1346,7 +1373,7 @@ exif_parse (struct jpeg_decompress_struct* cinfo,
     }
   }
 
-  return -1;
+  return 0;
 }
 
 /** libjpeg error manager */
@@ -1368,6 +1395,38 @@ exif_error_exit (struct jpeg_common_struct* cinfo)
   longjmp (((struct exif_error_mgr*) cinfo->err)->setjmp_buffer, 1);
 }
 
+/**
+ * Construct a file name
+ * @param base		the base file name
+ * @param baselen	length of the base file name
+ * @param dir		the directory name
+ * @param dirlen	length of the directory name
+ * @return		the constructed file name, or 0 on error
+ */
+static char*
+createfilename (const char* base,
+		unsigned baselen,
+		const char* dir,
+		unsigned dirlen)
+{
+  const char* s;
+  char* fnam;
+  for (s = base + baselen; s > base; s--)
+    if (*s == '/'
+#if defined WIN32 || defined __WIN32
+	|| *s == '\\' || *s == ':'
+#endif
+	) { s++; break; }
+  baselen -= s - base;
+  fnam = malloc (dirlen + baselen + 2);
+  if (fnam) {
+    memcpy (fnam, dir, dirlen);
+    fnam[dirlen] = '/';
+    memcpy (fnam + dirlen + 1, s, baselen + 1);
+  }
+  return fnam;
+}
+
 int
 main (int argc, char **argv)
 {
@@ -1380,14 +1439,22 @@ main (int argc, char **argv)
 
   /** flag: remove Exif IFD1 */
   int rmifd1 = 1;
-  /** image transformation (-1=according to the Orientation tag) */
-  int transformation = -1;
+  /** image transformation (0=according to the Orientation tag) */
+  enum JXFORM_CODE transformation = 0;
   /** time stamp for output files */
   enum { tdt, tdtDigitized, tdtOriginal, tdtNone } tOutput = tdt;
   /** flag: disable the processing of options */
   int ignopt = 0;
   /** flag: use the progressive JPEG process */
   int progressive = 1;
+  /** directory for modified image files (0=replace the original files) */
+  char* imagedir = 0;
+  /** directory for Exif thumbnail files (0=output disabled) */
+  char* thumbdir = 0;
+  /** imagedir length */
+  unsigned imagedirlen = 0;
+  /** thumbdir length */
+  unsigned thumbdirlen = 0;
 
   srcinfo.err = dstinfo.err = jpeg_std_error (&jerr.pub);
   jerr.pub.addon_message_table = exif_message_table;
@@ -1428,7 +1495,7 @@ main (int argc, char **argv)
 	return 1;
       case '?':
       case 'h':
-	fputs ("exifiron 1.0\n"
+	fputs ("exifiron 1.1\n"
 	       "Orientation correction and shrinking of Exif files\n"
 	       "Usage: exifiron [options] file ...\n"
 	       "Options:\n"
@@ -1464,9 +1531,13 @@ main (int argc, char **argv)
 	       "\tbaseline JPEG\n"
 	       "-p"
 	       "\tprogressive JPEG (default)\n", stderr);
+	fputs ("-s dir\n"
+	       "\tsave the image files in dir instead of replacing originals\n"
+	       "-S dir\n"
+	       "\tsave the Exif thumbnails in dir\n", stderr);
 	return 0;
 
-      case '0': transformation = -1; break;
+      case '0': transformation = 0; break;
       case '1': case 'u': transformation = JXFORM_NONE; break;
       case '2': case 'f': transformation = JXFORM_FLIP_H; break;
       case '3': case 'U': transformation = JXFORM_ROT_180; break;
@@ -1486,6 +1557,60 @@ main (int argc, char **argv)
 
       case 'b': progressive = 0; break;
       case 'p': progressive = 1; break;
+
+      case 's': case 'S':
+	if (argc <= 2) {
+	  putc ('-', stderr), putc (*opts, stderr);
+	  fputs (": missing argument (use -h for help)\n", stderr);
+	  return 1;
+	}
+	argc--;
+	if (!**++argv) {
+	  /* empty string argument => restore default behaviour */
+	  if (*opts == 's')
+	    imagedir = 0, imagedirlen = 0;
+	  else
+	    thumbdir = 0, thumbdirlen = 0;
+	}
+	else {
+	  unsigned namelen = strlen (*argv);
+#if defined WIN32 || defined __WIN32
+	  char* s = *argv;
+	  while (*s) if (*s++ == '\\') s[-1] = '/';
+#endif
+	  if (*opts == 's')
+	    imagedir = *argv, imagedirlen = namelen;
+	  else
+	    thumbdir = *argv, thumbdirlen = namelen;
+	  if (mkdir (*argv
+#if !defined WIN32 && !defined __WIN32
+		     , 0777
+#endif
+		     ) && errno != EEXIST) {
+	    fputs (*argv, stderr);
+	    perror (": mkdir");
+	    return 2;
+	  }
+	  else {
+	    struct stat statbuf;
+	    char* s = malloc (namelen + 3);
+	    if (!s) {
+	      putc ('-', stderr), putc (*opts, stderr);
+	      fputs (": out of memory\n", stderr);
+	      return 2;
+	    }
+	    memcpy (s, *argv, namelen);
+	    memcpy (s + namelen, "/.", 3);
+	    if (stat (s, &statbuf)) {
+	      fputs (s, stderr);
+	      perror (": stat");
+	      free (s);
+	      return 2;
+	    }
+	    free (s);
+	  }
+	}
+	break;
       }
     }
 
@@ -1519,10 +1644,15 @@ main (int argc, char **argv)
     jpeg_read_header (&srcinfo, TRUE);
 
     datetime = tOutput;
+    dstname = thumbdir
+      ? createfilename (*argv, strlen (*argv), thumbdir, thumbdirlen)
+      : 0;
     info.transform = exif_parse (&srcinfo, transformation,
 				 tOutput == tdtNone ? 0 : &datetime,
-				 rmifd1);
-    jfif = info.transform == -1;
+				 rmifd1, dstname);
+    if (dstname)
+      free (dstname), dstname = 0;
+    jfif = !info.transform;
     if (jfif)
       info.transform = JXFORM_NONE;
 
@@ -1541,23 +1671,35 @@ main (int argc, char **argv)
 						    src_coef_arrays,
 						    &info);
     if (!opts) {
-      /* allocate destination file name, change the last character to '0' */
-      unsigned len = strlen (*argv) + 1;
-      if (!(dstname = malloc (len))) {
+      int fd;
+      /* allocate the destination file name */
+      unsigned len = strlen (*argv);
+      if (!(dstname = imagedir
+	    ? createfilename (*argv, len, imagedir, imagedirlen)
+	    : malloc (len + 1))) {
 	fputs (*argv, stderr);
 	fputs (": out of memory\n", stderr);
 	goto quit;
       }
-      memcpy (dstname, *argv, len);
-      dstname[len - 2] = '0';
-      if (!(f = fopen (dstname, "wb"))) {
+      if (!imagedir) {
+	memcpy (dstname, *argv, len + 1);
+	dstname[len - 1] = '0'; /* change the last character to '0' */
+      }
+      unlink (dstname);
+      fd = open (dstname,
+#if defined WIN32 || defined __WIN32
+		 O_WRONLY | O_CREAT | O_EXCL | O_BINARY, 0666
+#else
+		 O_WRONLY | O_CREAT | O_EXCL, 0444
+#endif
+		 );
+      if (fd < 0 || !(f = fdopen (fd, "wb"))) {
 	fputs (dstname, stderr);
 	perror (": open (writing)");
 	free (dstname); dstname = 0;
 	goto quit;
       }
     }
-    /* to do: allocate destination file name, open destination file */
     jpeg_stdio_dest (&dstinfo, f);
     dstinfo.optimize_coding = TRUE;
     dstinfo.write_JFIF_header = jfif;
@@ -1570,17 +1712,9 @@ main (int argc, char **argv)
     jpeg_finish_compress (&dstinfo);
     if (!opts) {
       fclose (f);
-      if (datetime != tOutput) {
-	/* modify the file timestamp */
-	struct utimbuf utim;
-	utim.actime = utim.modtime = datetime;
-	if (utime (dstname, &utim)) {
-	  fputs (*argv, stderr);
-	  perror (": utime");
-	  /* non-fatal error, don't abort */
-	}
-      }
-      if (rename (dstname, *argv)) {
+      if (datetime != tOutput)
+	touchfile (dstname, datetime);
+      if (!imagedir && rename (dstname, *argv)) {
 	fputs (dstname, stderr);
 	fputs (" -> ", stderr);
 	fputs (*argv, stderr);
